# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Manage a MAAS installation."""

from __future__ import (
    absolute_import,
    print_function,
    unicode_literals,
    )

str = None

__metaclass__ = type
__all__ = [
    'MAASFixture',
    ]


import httplib
import json
import logging
from random import choice
from string import ascii_letters
from textwrap import dedent

from apiclient.creds import convert_string_to_tuple
from apiclient.maas_client import (
    MAASClient,
    MAASDispatcher,
    MAASOAuth,
    urllib2,
    )
from fixtures import Fixture
from maastest import utils
import netaddr
from testtools.monkey import MonkeyPatcher


MAAS_ADMIN_USER = 'admin'


def compose_rewrite_for_default_maas_url(config_file, default_maas_url):
    """Return a shell command to rewrite `DEFAULT_MAAS_URL` in config."""
    # Comment out existing DEFAULT_MAAS_URL, and add new one.
    replacement = dedent("""
        # Replaced by maas-test:
        # \\1
        DEFAULT_MAAS_URL = "%s"
        """ % default_maas_url).replace('\n', '\\n')

    return [
        'sed',
        '--in-place=.bak',
        's|^\\(DEFAULT_MAAS_URL *=.*\\)$|%s|' % replacement,
        config_file,
        ]


class MAASFixture(Fixture):
    """A fixture for a MAAS server."""

    def __init__(self, kvmfixture):
        """Initialise a MAAS server installed in a KVM instance.

        :param kvmfixture: The `maastest.KVMFixture` corresponding to the
            KVM instance on which the MAAS server will be installed.
        """
        self.kvm_fixture = kvmfixture

    def install_maas(self):
        """Install and configure MAAS in the virtual machine."""
        # Install the English language pack first.  If this is not done
        # before postgres is installed, on some systems it won't be able to
        # set up its main cluster.
        # TODO: Is there no better way to ensure this, e.g. through a
        # dependency?
        logging.info("Installing MAAS...")
        self.kvm_fixture.install_packages(['language-pack-en'])
        # Now we can install maas (which also installs postgres).
        self.kvm_fixture.install_packages(['maas', 'maas-dhcp', 'maas-dns'])
        logging.info("Done installing MAAS.")

    def configure_default_maas_url(self):
        """Set `DEFAULT_MAAS_URL` in the virtual machine.

        We'd prefer to do this by preseed, before installing MAAS, but that
        doesn't currently work (bug 1251175).  Instead, we patch up
        configuration after installation.
        """
        assert self.kvm_fixture.direct_ip is not None, (
            "configure_default_maas_url should only be called on a "
            "machine that has a 'direct interface.'")
        rewrite_command = compose_rewrite_for_default_maas_url(
            '/etc/maas/maas_local_settings.py',
            'http://%s/MAAS' % self.kvm_fixture.direct_ip)
        self.kvm_fixture.run_command(
            ['sudo'] + rewrite_command, check_call=True)
        self.kvm_fixture.run_command(
            ['sudo', 'service', 'apache2', 'restart'], check_call=True)

    def query_api_key(self, username):
        """Return the API key for the given MAAS user."""
        # The "apikey" command prints the user's API key to stdout.
        return_code, stdout, stderr = self.kvm_fixture.run_command([
            'sudo', 'maas', 'apikey',
            '--username=%s' % username,
            ], check_call=True)
        return stdout.strip()

    def create_maas_admin(self):
        """Create a MAAS admin user.

        Invoke this after MAAS has been installed on the virtual machine.

        :return: User name and password for the new admin.
        """
        username = MAAS_ADMIN_USER
        password = ''.join(choice(ascii_letters) for counter in range(8))
        # Email address really does not matter.  MAAS never sends anything
        # there. Its hostname doesn't have to resolve, although it has to
        # look like a FQDN.
        email = 'root@localhost.local'

        return_code, _, stderr = self.kvm_fixture.run_command([
            'sudo', 'maas', 'createadmin',
            '--username=%s' % username,
            '--password=%s' % password,
            '--email=%s' % email,
            ], check_call=True)

        return username, password

    def import_maas_images(self):
        """Import boot images into the MAAS instance.

        This download gigabytes of data to the virtual machine's disk.

        :param kvm_fixture: `KVMFixture` that controls the virtual machine.
        :param admin_user: Login name for the MAAS admin user.
        :param api_key: Admin user's API key.
        """
        logging.info("Importing MAAS PXE files...")
        # TODO: ARCHES and RELEASES should come from command-line
        # parameters.
        self.kvm_fixture.run_command([
            'sudo',
            'RELEASES=precise',
            'ARCHES=amd64/generic i386/generic',
            'maas-import-pxe-files',
            ],
            check_call=True)
        logging.info("Done importing MAAS PXE Files.")

    def setUp(self):
        """Install and configure MAAS inside the virtual machine.

        :param kvm_fixture: `KVMFixture` that controls the virtual machine.
        :return: Username, password and API key for the MAAS admin user.
        """
        super(MAASFixture, self).setUp()
        # MAAS installation and configuration.
        self.install_maas()
        if self.kvm_fixture.direct_ip is not None:
            self.configure_default_maas_url()
        # Admin creation.
        admin_user, admin_password = self.create_maas_admin()
        api_key = self.query_api_key(admin_user)
        maas_client = self.get_maas_api_client(api_key)
        self.admin_user = admin_user
        self.admin_password = admin_password
        self.admin_api_key = api_key
        self.admin_maas_client = maas_client
        # MAAS setup.
        self.import_maas_images()
        if self.kvm_fixture.direct_ip is not None:
            self.check_cluster_connected()
            self.configure_cluster()

    def get_maas_api_client(self, api_key):
        """Create and return a MAASClient.

        :param kvm_fixture: `KVMFixture` that controls the virtual machine.
        :return: `MAASClient` instance for the MAAS admin user.
        """
        credentials = convert_string_to_tuple(api_key)
        auth = MAASOAuth(*credentials)
        return MAASClient(
            auth, MAASDispatcher(),
            "http://%s/MAAS" % self.kvm_fixture.ip_address())

    def get_master_ng_uuid(self):
        """Get the UUID of the master nodegroup from the API."""
        uri = utils.get_uri('nodegroups/')
        response = self.admin_maas_client.get(uri, op='list')
        if response.code != httplib.OK:
            raise Exception("Error listing the clusters")
        nodegroups = json.loads(response.read())
        if len(nodegroups) != 1:
            raise Exception(
                "Expected exactly 1 nodegroup, but saw %d." % len(nodegroups))
        return nodegroups[0]['uuid']

    def check_cluster_connected(self):
        for retry in utils.retries(timeout=3 * 60):
            name = self.get_master_ng_uuid()
            if name != 'master':
                # The master cluster is connected.  The master nodegroup
                # had its uuid field updated from 'master' to its UUID.
                return
        raise Exception("Master cluster failed to connect.")

    def configure_cluster(self):
        network = self.kvm_fixture.direct_network
        first_ip = self.kvm_fixture.direct_first_available_ip()
        last_ip = "%s" % netaddr.IPAddress(
            self.kvm_fixture.direct_network.last - 1)
        dhcp_config = {
            "ip": self.kvm_fixture.direct_ip,
            "interface": "eth1",
            "subnet_mask": "%s" % network.netmask,
            "broadcast_ip": "%s" % network.broadcast,
            "router_ip": self.kvm_fixture.direct_ip,
            "management": '2',  # Manage DNS and DHCP.
            "ip_range_low": first_ip,
            "ip_range_high": last_ip,
        }
        uuid = self.get_master_ng_uuid()
        uri = utils.get_uri(
            'nodegroups/%s/interfaces/%s/' % (uuid, 'eth1'))

        # XXX: rvb 2013-11-14 bug=1251214: workaround a bug affecting
        # 'PUT' requests in urllib2 (which is used by MAASClient).
        patcher = MonkeyPatcher()
        patcher.add_patch(urllib2.Request, "get_method", lambda self: 'PUT')
        response = patcher.run_with_patches(
            self.admin_maas_client.put, uri, **dhcp_config)

        if response.code != httplib.OK:
            raise Exception("Error configure the master cluster")
