# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Test whether or not a node is compatible with MAAS."""

from __future__ import (
    absolute_import,
    print_function,
    unicode_literals,
    )

__metaclass__ = type
__all__ = [
    "main",
    ]

import httplib
import inspect
import json
import logging
import sys
import unittest

from maastest import utils
from maastest.kvmfixture import KVMFixture
from maastest.maasfixture import MAASFixture
import testresources
import testtools
from testtools.testcase import gather_details


class TestMAAS(testtools.TestCase, testresources.ResourcedTestCase):

    @classmethod
    def configure(cls, args):
        cls.args = args

    @classmethod
    def setUpClass(cls):
        super(TestMAAS, cls).setUpClass()
        # TODO: series and architecture should be script parameters.
        architecture = utils.determine_vm_architecture()
        cls.machine = KVMFixture(
            series='saucy', architecture=architecture,
            direct_interface=cls.args.interface)
        cls.machine.setUp()
        try:
            cls.maas = MAASFixture(cls.machine)
            cls.maas.setUp()
        except:
            cls.machine.cleanUp()
            raise

    @classmethod
    def tearDownClass(cls):
        cls.machine.cleanUp()
        super(TestMAAS, cls).tearDownClass()

    def shortDescription(self):
        """Overridable from `unittest`: Describe the ongoing test.

        By default this shows the first line of the test's docstring (if
        available), but this `TestCase` is basically our UI so we want full,
        user-friendly docstrings.
        """
        # Cribbed from unittest2.TestCase.  The difference is that we return
        # the whole docstring, instead of just the first line.
        # Future maintainers beware: the docstring may also be None.
        return self._testMethodDoc

    # XXX 2013-11-07 gmb bug=897190:
    #     This exists because of the above bug in testtools, which means
    #     that fixture details don't get included with the TestCase
    #     details on cleanUp().
    def unify_details(self):
        """Merge the details of the fixture into the details of the TestMAAS.
        """
        gather_details(self.machine.getDetails(), self.getDetails())

    def setUp(self):
        super(TestMAAS, self).setUp()
        self.addCleanup(self.unify_details)

    def print_connection_details(self):
        print(
            "\n"
            "URL: http://%s/MAAS/\n"
            "Username: %s\n"
            "Password: %s\n"
            "SSH login: ssh -i ~/.maas-test/vm_ssh_id_rsa %s\n"
            % (
                self.machine.ip_address(),
                self.maas.admin_user,
                self.maas.admin_password,
                self.machine.identity(),
            ))

    def get_node_list(self, status):
        """Return the list of nodes with the given status."""
        uri = utils.get_uri('nodes/')
        response = self.maas.admin_maas_client.get(uri, op="list")
        self.assertEqual(
            httplib.OK, response.code, "Failed to get the node list.")
        nodes = json.loads(response.read())
        relevant_nodes = [node for node in nodes if node['status'] == status]
        return relevant_nodes


class TestOneNode(TestMAAS):
    """Test if a node is compatible with MAAS."""

    def test_power_up_node(self):
        """Power up the node under test.

        If maas-test is running in interactive mode, this will ask the user
        to power up the node manually.  Otherwise it will use the provided
        power parameters to power up the node by running 'ipmipower' in the
        virtual machine..
        """
        # TODO: only print this in debug/verbose mode.
        self.print_connection_details()
        if self.args.interactive_enlistment:
            self.power_up_node_interactive()
        else:
            self.power_up_node_noninteractive()

    def power_up_node_interactive(self):
        """Asks the user to power up the node manually."""
        print(
            "Power up the node under test and press enter to proceed "
            "with the testing.")
        sys.stdin.readline()

    def power_up_node_noninteractive(self):
        """Power up the node using the provided power parameters."""
        # Scan the network to find the IP address of the node's BMC.
        for retry in utils.retries(delay=20, timeout=5 * 60):
            ip_scan = self.maas.kvm_fixture.get_ip_from_network_scan
            power_address = ip_scan(self.args.bmc_mac)
            if power_address is not None:
                break
        if power_address is None:
            self.fail("Failed to get the IP address of the BMC.")
        # Power-cycle the node.
        self.maas.kvm_fixture.run_command([
            'ipmipower', '-h', power_address,
            '-u', self.args.bmc_username,
            '-p', self.args.bmc_password, '--off'],
            check_call=True)
        self.maas.kvm_fixture.run_command([
            'ipmipower', '-h', power_address,
            '-u', self.args.bmc_username,
            '-p', self.args.bmc_password, '--on'],
            check_call=True)

    def test_enlist_node(self):
        """Enlist node."""
        for retry in utils.retries(delay=10, timeout=10 * 60):
            nb_enlisted_nodes = len(self.get_node_list(0))  # 0 = 'Declared'.
            if nb_enlisted_nodes == 1:
                return
        self.fail("Failed to enlist node.")

    def test_commission_node(self):
        """Commission node."""
        uri = utils.get_uri('nodes/')
        response = self.maas.admin_maas_client.post(uri, op="accept_all")
        self.assertEqual(
            httplib.OK, response.code, "Failed to accept node.")

        for retry in utils.retries(delay=10, timeout=10 * 60):
            nb_enlisted_nodes = len(self.get_node_list(4))  # 4 = 'Ready'.
            if nb_enlisted_nodes == 1:
                return
        self.fail("Failed to commission node.")


def check_kvm_ok():
    """Check that this machine is capable of running KVM.

    :return: True if KVM extensions are found, False otherwise.
    """
    logging.info("Checking for KVM extensions.")
    retcode, stdout, _ = utils.run_command(['sudo', 'kvm-ok'])
    if retcode != 0:
        logging.debug(stdout.strip())
        return False
    else:
        return True


def main(argv=None):
    from maastest import parser
    maas_test_parser = parser.prepare_parser()
    args = maas_test_parser.parse_args()
    output_stream = sys.stderr
    logging.basicConfig(
        level=logging.DEBUG, stream=output_stream, format="%(message)s")

    if not check_kvm_ok():
        logging.error("Unable to continue without KVM extensions.")
        return 1

    class ConfiguredTestMAAS(TestOneNode):
        """A configured version of TestInteractiveOneNode.

        ConfiguredTestMAAS is a TestMAAS use to configure it by calling
        cls.configure.  We need to configure the class itself and not
        the instance because the KVMFixture TestMAAS instanciates and needs
        to configure is created at the class level.
        """
    ConfiguredTestMAAS.configure(args)

    # Create a TestLoader which loads the tests in the order in which
    # they are defined in the test case class.
    loader = unittest.TestLoader()
    # 'method_starts' is a mapping between the test method names and the line
    # at which their definition starts in the source code.
    method_starts = dict(
        (name, inspect.getsourcelines(method)[1]) for (name, method)
        in inspect.getmembers(ConfiguredTestMAAS, predicate=inspect.ismethod)
        if name.startswith('test_'))
    loader.sortTestMethodsUsing = (
        lambda x, y: method_starts.get(x) - method_starts.get(y))

    suite = loader.loadTestsFromTestCase(ConfiguredTestMAAS)
    runner = unittest.TextTestRunner(
        stream=output_stream, verbosity=2, descriptions=False,
        failfast=True)
    runner.run(suite)


if __name__ == '__main__':
    returncode = main()
    if returncode is not None:
        sys.exit(returncode)
