# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Tests for the "prepare" command."""

from __future__ import (
    absolute_import,
    print_function,
    unicode_literals,
    )

str = None

__metaclass__ = type
__all__ = []

import httplib
import json
import os.path
from random import randint
from subprocess import check_call
from textwrap import dedent

from apiclient.maas_client import MAASClient
from fixtures import TempDir
from maastest import (
    kvmfixture,
    maasfixture,
    utils,
    )
import mock
import netaddr
from testtools import TestCase
from testtools.matchers import FileContains


class FakeResponse(object):
    """A fake http response used in testing."""

    def __init__(self, content, code):
        self.content = content
        self._code = code

    def read(self):
        return self.content

    @property
    def code(self):
        return self._code


class TestMAASFixture(TestCase):

    def make_kvm_fixture(self):
        """Create a `KVMFixture` with arbitrary series/architecture.

        Its `run_command` method will be replaced by a `MagicMock`, which
        returns success by default.  Its `ip_address` will be patched to return
        an arbitrary IP address.
        """
        fixture = kvmfixture.KVMFixture('droopy', 'i860')
        fixture.run_command = mock.MagicMock(
            return_value=(0, 'stdout output', 'stderr output'))
        ip_address = '.'.join('%d' % randint(1, 254) for octets in range(4))
        fixture.ip_address = mock.MagicMock(return_value=ip_address)
        return fixture

    def make_maas_local_settings(self, default_maas_url):
        """Compose a `maas_local_settings.py`, as installed."""
        return ('DEFAULT_MAAS_URL = "%s"' % default_maas_url).encode('ascii')

    def make_rewritten_maas_local_settings(self, old_url, new_url):
        """Compose a `maas_local_settings.py` as rewritten.

        We rewrite this file to change its `DEFAULT_MAAS_URL` setting.  It's
        done in a shell command, so needs careful testing.
        """
        return dedent("""
            # Replaced by maas-test:
            # DEFAULT_MAAS_URL = "%s"
            DEFAULT_MAAS_URL = "%s"
            """ % (old_url, new_url)).encode('ascii')

    def make_file(self, name=None, content=None):
        """Write a file."""
        self.assertIsInstance(content, bytes)
        if name is None:
            name = self.getUniqueString()
        path = os.path.join(self.useFixture(TempDir()).path, name)
        with open(path, 'wb') as f:
            f.write(content)
        return path

    def patch_MAASClient(self, client, method, content, code=httplib.OK):
        """Patch a MAASClient so that it will return a pre-cooked response."""
        json_content = json.dumps(content)
        response = FakeResponse(json_content, code)
        mock_object = mock.MagicMock(return_value=response)
        self.patch(client, method, mock_object)
        return mock_object

    def test_compose_rewrite_for_default_maas_url_rewrites_url(self):
        old_url = 'http://1.1.1.1/'
        new_url = 'http://2.2.2.2/'
        config_file = self.make_file(
            content=self.make_maas_local_settings(old_url))

        command = maasfixture.compose_rewrite_for_default_maas_url(
            config_file, new_url)
        # Test the generated sed command, by running it locally.  In real use
        # this runs only on the virtual machine.
        # (But make sure it doesn't sudo.  That won't do in tests.)
        self.assertNotIn('sudo', command)
        check_call(command)

        new_content = self.make_rewritten_maas_local_settings(old_url, new_url)
        self.assertThat(config_file, FileContains(new_content.decode('ascii')))

    def test_compose_rewrite_for_default_maas_url_accepts_path(self):
        old_url = 'http://old-host.com/OLDMAAS'
        new_url = 'http://new-host.local/NEWMAAS'
        config_file = self.make_file(
            content=self.make_maas_local_settings(old_url))

        check_call(
            maasfixture.compose_rewrite_for_default_maas_url(
                config_file, new_url))

        new_content = self.make_rewritten_maas_local_settings(old_url, new_url)
        self.assertThat(config_file, FileContains(new_content.decode('ascii')))

    def test_compose_rewrite_for_default_maas_url_preserves_other_text(self):
        top = b"# Text before setting."
        bottom = b"# Text after setting."
        old_url = "http://1.1.1.1/"
        old_content = b'\n'.join(
            [top, self.make_maas_local_settings(old_url), bottom])
        config_file = self.make_file(content=old_content)
        new_url = "http://2.2.2.2/"

        check_call(
            maasfixture.compose_rewrite_for_default_maas_url(
                config_file, new_url))

        new_part = self.make_rewritten_maas_local_settings(old_url, new_url)
        new_content = b'\n'.join([top, new_part, bottom])
        self.assertThat(config_file, FileContains(new_content.decode('ascii')))

    def test_compose_rewrite_for_default_maas_url_ignores_commented_URL(self):
        top = b'#DEFAULT_MAAS_URL = "http://9.9.9.9/"\n'
        old_url = "http://1.1.1.1/"
        old_content = top + self.make_maas_local_settings(old_url)
        config_file = self.make_file(content=old_content)
        new_url = "http://2.2.2.2/"

        check_call(
            maasfixture.compose_rewrite_for_default_maas_url(
                config_file, new_url))

        new_content = top + self.make_rewritten_maas_local_settings(
            old_url, new_url)
        self.assertThat(config_file, FileContains(new_content.decode('ascii')))

    def test_configure_default_maas_url_sets_config_and_restarts(self):
        self.patch(kvmfixture, 'run_command', mock.MagicMock())
        fixture = self.make_kvm_fixture()
        fixture.direct_ip = '127.99.88.77'
        rewrite_command = maasfixture.compose_rewrite_for_default_maas_url(
            '/etc/maas/maas_local_settings.py',
            "http://%s/MAAS" % fixture.direct_ip)

        maas_fixture = maasfixture.MAASFixture(fixture)
        maas_fixture.configure_default_maas_url()

        self.assertEqual([
            mock.call(['sudo'] + rewrite_command, check_call=True),
            mock.call(
                ['sudo', 'service', 'apache2', 'restart'], check_call=True),
            ],
            fixture.run_command.mock_calls)

    def test_create_maas_admin_creates_admin(self):
        fixture = self.make_kvm_fixture()

        maas_fixture = maasfixture.MAASFixture(fixture)
        username, password = maas_fixture.create_maas_admin()

        create_call = fixture.run_command.call_args_list[0]
        args, kwargs = create_call
        command_line = args[0]
        self.assertEqual(
            ['sudo', 'maas', 'createadmin'],
            command_line[:3])
        self.assertIn('--username=%s' % username, command_line)
        self.assertIn('--password=%s' % password, command_line)

    def test_query_api_key_returns_api_key(self):
        username = "maas-user"
        actual_api_key = "API KEY"
        fixture = self.make_kvm_fixture()
        fixture.run_command.return_value = (0, actual_api_key, 'stderr output')

        maas_fixture = maasfixture.MAASFixture(fixture)
        api_key = maas_fixture.query_api_key(username)

        args, kwargs = fixture.run_command.call_args_list[0]
        command_line = args[0]
        self.assertEqual(
            ['sudo', 'maas', 'apikey', '--username=%s' % username],
            command_line)
        self.assertEqual(actual_api_key, api_key)

    def test_get_maas_api_client_returns_client(self):
        actual_api_key = "actual:maas:key"
        self.patch(kvmfixture, 'run_command', mock.MagicMock(
            return_value=(0, b'stdout output', b'stderr output')))
        fixture = self.make_kvm_fixture()
        fixture.run_command.return_value = (
            0, actual_api_key, b'stderr output')
        maas_fixture = maasfixture.MAASFixture(fixture)
        client = maas_fixture.get_maas_api_client(actual_api_key)
        self.assertIsInstance(client, MAASClient)

    def test_install_maas_installs_integrates_with_install_packages(self):
        fixture = self.make_kvm_fixture()
        maas_fixture = maasfixture.MAASFixture(fixture)

        maas_fixture.install_maas()

        fixture.run_command.assert_has_call(
            mock.call(
                kvmfixture.APT_GET_INSTALL + ['language-pack-en'],
                check_call=True),
            )

    def test_install_maas_installs_language_pack_followed_by_maas(self):
        # We need to do it in this order, or postgres (and other things) may
        # break during installation.
        fixture = self.make_kvm_fixture()
        fake_install = mock.MagicMock()
        maas_fixture = maasfixture.MAASFixture(fixture)
        self.patch(fixture, 'install_packages', fake_install)

        maas_fixture.install_maas()

        self.assertEqual(
            [
                mock.call(['language-pack-en']),
                mock.call(['maas', 'maas-dhcp', 'maas-dns']),
            ],
            fake_install.mock_calls)

    def test_get_master_ng_uuid_returns_uuid(self):
        fixture = self.make_kvm_fixture()
        maas_fixture = maasfixture.MAASFixture(fixture)
        self.patch(maas_fixture, 'install_packages', mock.MagicMock())
        self.patch(maas_fixture, 'admin_maas_client', mock.MagicMock())

        uuid = self.getUniqueString()
        response = [{'uuid': uuid}]
        call_mock = self.patch_MAASClient(
            maas_fixture.admin_maas_client, 'get', response)

        self.assertEqual(
            (uuid, [mock.call('/api/1.0/nodegroups/', op='list')]),
            (maas_fixture.get_master_ng_uuid(), call_mock.mock_calls))

    def test_check_cluster_connected_returns_if_cluster_connected(self):
        fixture = self.make_kvm_fixture()
        maas_fixture = maasfixture.MAASFixture(fixture)
        self.patch(maas_fixture, 'install_packages', mock.MagicMock())
        self.patch(maas_fixture, 'admin_maas_client', mock.MagicMock())

        uuid = self.getUniqueString()
        response = [{'uuid': uuid}]
        self.patch_MAASClient(
            maas_fixture.admin_maas_client, 'get', response)

        self.assertEqual(None, maas_fixture.check_cluster_connected())

    def test_check_cluster_connected_fails_if_cluster_not_connected(self):
        fixture = self.make_kvm_fixture()
        maas_fixture = maasfixture.MAASFixture(fixture)
        self.patch(maas_fixture, 'install_packages', mock.MagicMock())
        self.patch(maas_fixture, 'admin_maas_client', mock.MagicMock())
        self.patch(
            utils, 'retries', mock.MagicMock(return_value=[(0, 1)]))

        response = [{'uuid': 'master'}]
        call_mock = self.patch_MAASClient(
            maas_fixture.admin_maas_client, 'get', response)

        self.assertRaises(Exception, maas_fixture.check_cluster_connected)
        self.assertEqual(
            [mock.call('/api/1.0/nodegroups/', op='list')],
            call_mock.mock_calls)

    def test_configure_cluster_updates_dhcp_info(self):
        fixture = self.make_kvm_fixture()
        fixture.direct_ip = '192.168.2.0'
        network = netaddr.IPNetwork('192.168.2.0/24')
        fixture.direct_network = network
        maas_fixture = maasfixture.MAASFixture(fixture)
        self.patch(maas_fixture, 'install_packages', mock.MagicMock())
        self.patch(maas_fixture, 'admin_maas_client', mock.MagicMock())
        self.patch(
            utils, 'retries', mock.MagicMock(return_value=[(0, 1)]))

        uuid = self.getUniqueString()
        response = [{'uuid': uuid}]
        call_mock = self.patch_MAASClient(
            maas_fixture.admin_maas_client, 'get', response)
        put_mock = self.patch_MAASClient(
            maas_fixture.admin_maas_client, 'put', '')

        maas_fixture.configure_cluster()
        interface_url = '/api/1.0/nodegroups/%s/interfaces/%s/' % (
            uuid, 'eth1')
        last_ip = "%s" % netaddr.IPAddress(
            fixture.direct_network.last - 1)
        self.assertEqual(
            (
                [mock.call('/api/1.0/nodegroups/', op='list')],
                [mock.call(
                    interface_url,
                    ip=fixture.direct_ip,
                    interface="eth1",
                    subnet_mask="%s" % network.netmask,
                    broadcast_ip="%s" % network.broadcast,
                    router_ip=fixture.direct_ip,
                    management='2',
                    ip_range_low=fixture.direct_first_available_ip(),
                    ip_range_high=last_ip,
                    )],
            ),
            (call_mock.mock_calls, put_mock.mock_calls))

    def test_setUp_sets_admin_details(self):
        fixture = self.make_kvm_fixture()
        maas_fixture = maasfixture.MAASFixture(fixture)
        fixture.direct_ip = None
        user = 'maas-admin-user'
        password = 'maas-admin-password'
        self.patch(
            maas_fixture, 'create_maas_admin',
            mock.MagicMock(return_value=(user, password)))
        key = 'fake:api:key'
        self.patch(
            maas_fixture, 'query_api_key', mock.MagicMock(return_value=key))

        with maas_fixture:
            pass

        client_key = "%s:%s:%s" % (
            maas_fixture.admin_maas_client.auth.consumer_token.key,
            maas_fixture.admin_maas_client.auth.resource_token.key,
            maas_fixture.admin_maas_client.auth.resource_token.secret)
        self.assertEqual(
            (user, password, key, key),
            (
                maas_fixture.admin_user,
                maas_fixture.admin_password,
                maas_fixture.admin_api_key,
                client_key,
            ))

    def test_setUp_calls_methods_if_direct_ip_set(self):
        fixture = self.make_kvm_fixture()
        fixture.direct_ip = '127.11.24.38'
        maas_fixture = maasfixture.MAASFixture(fixture)
        self.patch(
            maas_fixture, 'configure_cluster', mock.MagicMock())
        self.patch(
            maas_fixture, 'check_cluster_connected', mock.MagicMock())
        self.patch(
            maas_fixture, 'query_api_key',
            mock.MagicMock(return_value='fake:api:key'))
        # Patch methods that we want to check setUp calls.
        self.patch(
            maas_fixture, 'configure_default_maas_url', mock.MagicMock())
        self.patch(
            maas_fixture, 'check_cluster_connected', mock.MagicMock())
        self.patch(
            maas_fixture, 'configure_cluster', mock.MagicMock())

        with maas_fixture:
            pass

        self.assertEqual(
            [[mock.call()], [mock.call()], [mock.call()]],
            [
                maas_fixture.configure_default_maas_url.mock_calls,
                maas_fixture.check_cluster_connected.mock_calls,
                maas_fixture.configure_cluster.mock_calls,
            ])

    def test_setUp_does_not_call_methods_if_direct_ip_unset(self):
        fixture = self.make_kvm_fixture()
        maas_fixture = maasfixture.MAASFixture(fixture)
        self.patch(
            maas_fixture, 'query_api_key',
            mock.MagicMock(return_value='fake:api:key'))
        # Patch methods that we want to check setUp doesn't call.
        self.patch(
            maas_fixture, 'configure_default_maas_url', mock.MagicMock())
        self.patch(
            maas_fixture, 'check_cluster_connected', mock.MagicMock())
        self.patch(
            maas_fixture, 'configure_cluster', mock.MagicMock())

        with maas_fixture:
            pass

        self.assertEqual(
            [[], [], []],
            [
                maas_fixture.configure_default_maas_url.mock_calls,
                maas_fixture.check_cluster_connected.mock_calls,
                maas_fixture.configure_cluster.mock_calls,
            ])

    def test_import_mmas_images_imports_pxe_images(self):
        kvm_fixture = self.make_kvm_fixture()
        self.patch(kvm_fixture, 'run_command', mock.MagicMock())
        maas_fixture = maasfixture.MAASFixture(kvm_fixture)
        maas_fixture.import_maas_images()
        self.assertEqual(
            mock.call([
                u'sudo',
                u'RELEASES=precise',
                u'ARCHES=amd64/generic i386/generic',
                u'maas-import-pxe-files',
            ], check_call=True),
            kvm_fixture.run_command.mock_calls[-1])
