# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Tests for `maastest.utils`."""

from __future__ import (
    absolute_import,
    print_function,
    unicode_literals,
    )

__metaclass__ = type
__all__ = []

import os.path
import platform
from subprocess import PIPE
from textwrap import dedent
import time

import distro_info
from fixtures import TempDir
from maastest import utils
from maastest.utils import binary_content
import mock
from six import text_type
import testtools
from testtools.matchers import (
    MatchesRegex,
    MatchesStructure,
    )


class TestRunCommand(testtools.TestCase):

    def test_read_file_returns_file_contents_as_bytes(self):
        temp_dir = self.useFixture(TempDir())
        sample_file = os.path.join(temp_dir.path, self.getUniqueString())
        contents = self.getUniqueString().encode('ascii')
        with open(sample_file, 'wb') as f:
            f.write(contents)

        read_contents = utils.read_file(sample_file)

        self.assertEqual(contents, read_contents)
        self.assertIsInstance(contents, bytes)

    def test_run_command_calls_Popen(self):
        mock_Popen = mock.MagicMock()
        expected_retcode = self.getUniqueInteger()
        expected_stdout = self.getUniqueString()
        expected_stderr = self.getUniqueString()
        mock_Popen.return_value.returncode = expected_retcode
        mock_Popen.return_value.communicate.return_value = (
            expected_stdout, expected_stderr)
        self.patch(utils, 'Popen', mock_Popen)
        args = ['one', 'two']

        retcode, stdout, stderr = utils.run_command(args)

        self.assertEqual(
            [
                mock.call(
                    ['one', 'two'], stdout=PIPE, stderr=PIPE,
                    stdin=PIPE, shell=False),
                mock.call().communicate(None),
            ],
            mock_Popen.mock_calls)
        self.assertEqual(
            (expected_retcode, expected_stdout, expected_stderr),
            (retcode, stdout, stderr))

    def test_run_command_runs_command(self):
        retcode, stdout, stderr = utils.run_command(['ls', '/'])

        self.assertEqual((0, b''), (retcode, stderr))
        self.assertIn(b"boot", stdout)

    def test_run_command_checks_return_value(self):
        mock_Popen = mock.MagicMock()
        expected_retcode = 2
        expected_stdout = self.getUniqueString()
        expected_stderr = self.getUniqueString()
        mock_Popen.return_value.returncode = expected_retcode
        mock_Popen.return_value.communicate.return_value = (
            expected_stdout, expected_stderr)
        self.patch(utils, 'Popen', mock_Popen)
        args = ['one', 'two']

        error = self.assertRaises(
            Exception, utils.run_command, args, check_call=True)
        self.assertIn(expected_stdout, text_type(error))
        self.assertIn(expected_stderr, text_type(error))

    def test_run_command_uses_input(self):
        input_string = self.getUniqueString().encode("ascii")
        retcode, stdout, stderr = utils.run_command(
            ['cat', '-'], input=input_string)

        self.assertEqual(
            (0, b'', input_string), (retcode, stderr, stdout))


class TestBinaryContent(testtools.TestCase):
    """Tests for `binary_content`."""

    def test_returns_same_content(self):
        content = binary_content(b'abc123')
        self.assertEqual(
            b''.join(content.iter_bytes()),
            b'abc123')

    def test_has_appropriate_content_type(self):
        content = binary_content(b'abc123')
        self.assertThat(
            content.content_type,
            MatchesStructure.byEquality(
                type="application",
                subtype="octet-stream",
                parameters={},
            ))


class TestGetURI(testtools.TestCase):
    """Tests for `get_uri`."""

    def test_returns_api_root_plus_path(self):
        path = "this/is/a/path"
        self.assertEqual("/api/1.0/" + path, utils.get_uri(path))


class TestRetries(testtools.TestCase):

    def test_returns_retry_iterator(self):
        # Patch utils.sleep() so that no time is actually spent
        # sleeping.
        mock_sleep = mock.MagicMock()
        self.patch(utils, 'sleep', mock_sleep)
        # Patch utils.time() so that it will return [0, 0, 1, 2, ...] thus
        # simulating a sleep() of one second during each loop.  The
        # double '0' at the beginning of the list is there to cope with
        # retries() calling time() twice before the first iteration
        # starts.
        mock_time = mock.MagicMock()
        values = iter([0] + range(20))
        mock_time.side_effect = lambda: values.next()
        self.patch(utils, 'time', mock_time)
        timeout = 20
        tries = utils.retries(timeout=timeout)
        self.assertEquals(
            (timeout, timeout),
            (len(list(tries)), len(mock_sleep.mock_calls)))

    def test_returns_immediately(self):
        # When iterating over a generator returned by retries(), the
        # first object is returned immediately; sleep() is only called
        # between iterations.
        delay = 10
        tries = utils.retries(delay=delay)
        start = time.time()
        for _ in tries:
            elapsed = time.time() - start
            break
        self.assertLess(elapsed, delay)


class TestDetermineVMArchitecture(testtools.TestCase):

    def test_determine_vm_architecture_gets_system_arch(self):
        # Additions may become necessary.
        supported_architectures = {
            'amd64',
            'armhf',
            'i386',
        }
        self.assertIn(
            utils.determine_vm_architecture(),
            supported_architectures)

    def test_determine_vm_architecture_returns_ubuntu_arch_name(self):
        # Architecture names according to Python's platform module, and their
        # Ubuntu equivalents.  Unknown names are passed on unchanged.
        arch_names = {
            'i386': 'i386',
            'i686': 'i386',
            'x86_64': 'amd64',
            'armhf': 'armhf',
            'power': 'power',
            'sparc64': 'sparc64',
            'm68k': 'm68k',
        }
        # See what architectures determine_vm_architecture() reports for these
        # systems.
        translations = {}
        machine_func = mock.MagicMock()
        self.patch(platform, 'machine', machine_func)
        for arch in arch_names.keys():
            machine_func.return_value = arch
            translations[arch] = utils.determine_vm_architecture()

        self.assertEqual(arch_names, translations)

    def test_determine_vm_architecture_reports_failure(self):
        # platform.machine() returns an empty string when it can't determine
        # the architecture.  This test simulates that scenario.
        # In python 3 it returns a unicode string, in python 2 a string of
        # bytes.
        return_type = platform.machine().__class__
        self.patch(platform, 'machine', lambda: return_type())

        self.assertRaises(
            utils.UnknownCPUArchitecture,
            utils.determine_vm_architecture)


class TestDetermineVMSeries(testtools.TestCase):

    def test_determine_vm_series_returns_system_series(self):
        # We don't know the names of future series this test might run on, but
        # it'll be a lower-case name.  The supported releases start with
        # Precise, and for the forseeable future, progress through the
        # alphabet from there.
        self.assertThat(
            utils.determine_vm_series(),
            MatchesRegex('[p-z][a-z]+$'))

    def test_determine_vm_series_uses_distro_information(self):
        series = 'nerdy'
        distro = ('Ubuntu', '1.01', series)
        self.patch(platform, 'linux_distribution', lambda: distro)

        self.assertEqual(series, utils.determine_vm_series())


class TestGetSupportedSeries(testtools.TestCase):

    def test_get_supported_series_return_series(self):
        supported = (
            distro_info.UbuntuDistroInfo().supported(result="codename"))
        # The list of supported series will evolve in time.  At the time
        # of this writing, 'lucid' is still a supported series but it's
        # not supported by MAAS.
        if 'lucid' in supported:
            supported.remove('lucid')
            self.assertEqual(supported, utils.get_supported_series())
        else:
            self.assertEqual(supported, utils.get_supported_series())


class TestExtractMAPIPMapping(testtools.TestCase):

    def test_extract_mac_ip_mapping_parses_output(self):
        NMAP_XML_OUTPUT = dedent("""
            <nmaprun scanner="nmap" args="nmap -sP -oX - 192.168.2.0/24">
              <host>
                <address addr="192.168.2.2" addrtype="ipv4"/>
                <address addr="00:9C:02:A2:82:74" addrtype="mac"/>
              </host>
              <host>
                <address addr="192.168.2.4" addrtype="ipv4"/>
                <address addr="00:9C:02:A0:4D:0A" addrtype="mac"/>
              </host>
            </nmaprun>
        """)
        expected_result = {
            '00:9C:02:A2:82:74': '192.168.2.2',
            '00:9C:02:A0:4D:0A': '192.168.2.4',
        }
        self.assertEqual(
            expected_result, utils.extract_mac_ip_mapping(NMAP_XML_OUTPUT))

    def test_extract_mac_ip_mapping_returns_uppercase_mac(self):
        NMAP_XML_OUTPUT = dedent("""
            <nmaprun scanner="nmap" args="nmap -sP -oX - 192.168.2.0/24">
              <host>
                <address addr="192.168.2.2" addrtype="ipv4"/>
                <address addr="AA:bb:cc:dd:ee:FF" addrtype="mac"/>
              </host>
            </nmaprun>
        """)
        expected_result = {'AA:BB:CC:DD:EE:FF': '192.168.2.2'}
        self.assertEqual(
            expected_result, utils.extract_mac_ip_mapping(NMAP_XML_OUTPUT))

    def test_extract_mac_ip_mapping_with_empty_doc(self):
        self.assertEqual(
            {}, utils.extract_mac_ip_mapping('<nmaprun></nmaprun>'))

    def test_extract_mac_ip_mapping_with_missing_info(self):
        # If either the IP address of the MAC address is missing in a host
        # definition, the host entry is not considered.
        NMAP_XML_OUTPUT = dedent("""
            <nmaprun scanner="nmap" args="nmap -sP -oX - 192.168.2.0/24">
              <host>
                <address addr="192.168.2.2" addrtype="ipv4"/>
              </host>
              <host>
                <address addr="00:9C:02:A0:4D:0A" addrtype="mac"/>
              </host>
            </nmaprun>
        """)
        self.assertEqual({}, utils.extract_mac_ip_mapping(NMAP_XML_OUTPUT))
