# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""maas-test utilities."""

from __future__ import (
    absolute_import,
    print_function,
    unicode_literals,
    )

__metaclass__ = type
__all__ = [
    "BINARY",
    "binary_content",
    'determine_vm_series',
    'determine_vm_architecture',
    "get_uri",
    "read_file",
    "retries",
    "run_command",
    ]

from pipes import quote
import platform
from subprocess import (
    PIPE,
    Popen,
    )
from time import (
    sleep,
    time,
    )

import distro_info
from lxml import etree
from testtools.content import Content
from testtools.content_type import ContentType


def read_file(path):
    """Return a given file's contents, as `bytes`."""
    with open(path, "rb") as f:
        return f.read()


def run_command(args, input=None, check_call=False):
    """A wrapper to Popen to run commands in the command-line."""
    process = Popen(args, stdout=PIPE, stderr=PIPE, stdin=PIPE, shell=False)
    stdout, stderr = process.communicate(input)
    retcode = process.returncode
    if check_call and retcode != 0:
        cmd = " ".join(quote(arg) for arg in args)
        raise Exception(
            "Command '%s' failed (%d):\n%s\n%s" % (
                cmd, retcode, stdout, stderr))
    return retcode, stdout, stderr


def get_uri(path):
    """GET an API V1 URI.

    :return: The API URI.
    """
    api_root = '/api/1.0/'
    return api_root + path

BINARY = ContentType("application", "octet-stream")


def binary_content(data):
    """Create a `Content` object from a byte string.

    This is useful for adding details which are blobs of binary data.
    """
    return Content(BINARY, lambda: [data])


def retries(timeout=30, delay=1):
    """Helper for retrying something, sleeping between attempts.

    Yields ``(elapsed, remaining)`` tuples, giving times in seconds.

    @param timeout: From now, how long to keep iterating, in seconds.
    @param delay: The sleep between each iteration, in seconds.
    """
    start = time()
    end = start + timeout
    for now in iter(time, None):
        if now < end:
            yield now - start, end - now
            sleep(min(delay, end - now))
        else:
            break


class UnknownCPUArchitecture(RuntimeError):
    """Could not determine CPU architecture."""


def determine_vm_architecture():
    """Figure out CPU architecture for the virtual machine that will run MAAS.

    :return: Architecture string, e.g. 'i386'.
    :raise UnknownCPUArchitecture: if the CPU architecture could not be
        determined.
    """
    # Ubuntu sometimes uses different architecture names than we get from
    # the platform module.  This dict translates for those cases.
    ubuntu_names = {
        'i686': 'i386',
        'x86_64': 'amd64',
    }
    raw_arch = platform.machine()
    if raw_arch == '':
        raise UnknownCPUArchitecture(
            "Could not determine system CPU architecture.  "
            "Please report a bug against http://launchpad.net/maas-test "
            "with this system's details.  To work around the problem, "
            "try the --vm-arch option.")
    return ubuntu_names.get(raw_arch, raw_arch)


def determine_vm_series():
    """Figure out the Ubuntu release series to run MAAS on.

    :return: Series codename, e.g. "precise" for 12.04 Precise Pangolin.
    """
    distro, version, series = platform.linux_distribution()
    return series


def get_supported_series():
    """List of MAAS-supported Ubuntu series."""
    udi = distro_info.UbuntuDistroInfo()
    supported = udi.supported(result="codename")
    cut_off_index = (
        supported.index('precise') if 'precise' in supported else 0)
    return supported[cut_off_index:]


def extract_mac_ip_mapping(nmap_xml_scan):
    """Extract the IP->MAC address mapping from the result of a nmap scan.

    The returned mapping uses uppercase MAC addresses.

    :param nmap_xml_scan: The XML result of scanning a network using nmap.
        Scanning a network using nmap is typically done by running:
        `nmap -sP <network_definition> -oX -`.
    :type nmap_xml_scan: string
    """
    parser = etree.XMLParser(remove_blank_text=True)
    xml_doc = etree.fromstring(nmap_xml_scan, parser)
    hosts = xml_doc.xpath("/nmaprun/host")
    mapping = {}
    for host in hosts:
        ips = host.xpath("address[@addrtype='ipv4']/@addr")
        macs = host.xpath("address[@addrtype='mac']/@addr")
        if len(ips) == 1 and len(macs) == 1:
            mapping[macs[0].upper()] = ips[0]
    return mapping
